//======================================================== IBJdmn.c
//
// Data Manager for IBJ-programs
// =============================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-07-31  lj
// 2000-07-27 uj  extended for IBJodor
//
//==================================================================

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>

#include "IBJtxt.h"
#include "IBJmsg.h"
#include "IBJary.h"
#include "IBJdmn.h"

#define  MAXLINE  4000
#define  VERSION  1.2
char     *DmnVersion = "1.2.a";

#define  MAXNML  40
#define  STDEXT  ".dmna"
#define  BINEXT  ".dmnb"
#define  TXTEXT  ".dmnt"
#define  LISTSEP ";"
#define  STDFRM  "\t%11.3e"
#define  CR      "\r"
#define  LF      "\n"
#define  CRLF    "\r\n"
#define  SEPCHR  " ;\t"

#define  CHECK(a)  printf("check-%s\n", #a);  getchar();

enum DAT_TYPE { noDAT, chDAT, siDAT, inDAT, flDAT, tmDAT, ltDAT, lfDAT };

typedef struct frmrec {
  char name[40];
  enum DAT_TYPE dt;
  int ld, lf;
  char opt[40], iform[40], oform[40];
  float fac;
  char fspc;
  } DMNFRMREC;

static char *eMODn = "IBJdmn";
static char *DimParms =
  "lowb,hghb,size,dims,fact,mode,file,sequ,form,lsbf";

static int DatLen[] = { 0, 1, 2, 4, 4, 4, 8, 8 };
static DMNFRMREC *FrmTab;

/*=================================== Time ===============================*/

static long TmMax( void )
;
static long TmMin( void )
;
static char *Tmstring( long* )
;
static long TmValue( char* )
;
static void TimeStr( long, char* )
;
static long Seconds( char*, long )
;

static const char TmMinString[] = "-inf";
static const char TmMaxString[] = "+inf";
static const long TmMinValue = LONG_MIN;
static const long TmMaxValue = LONG_MAX;

  /*================================================================== TmMax
  */
static long TmMax( void )  /* return maximum value of time */
  {
  return TmMaxValue; }

  /*================================================================== TmMin
  */
static long TmMin( void )   /* return minimum value of time */
  {
  return TmMinValue; }

  /*=============================================================== TmString
  */
static char *TmString(     /* return string representation of time t */
  long *pt )               /* pointer to time t                      */
  {
  static char tstr[40];
  if (!pt)  strcpy( tstr, "NOT");
  else if (*pt == TmMinValue)  strcpy( tstr, TmMinString );
  else if (*pt == TmMaxValue)  strcpy( tstr, TmMaxString );
  else  TimeStr( *pt, tstr );
  return tstr;
  }

  /*================================================================ TmValue
  */
static long TmValue(        /* return binary representation of time */
  char *ts )                /* string representation of time        */
  {
  if (!strcmp(ts,TmMinString))  return TmMinValue;
  if (!strcmp(ts,TmMaxString))  return TmMaxValue;
  if ((!strcmp(ts,"UNDEF")) || (!strcmp(ts,"undef")) || (*ts == '?'))
    return TmMinValue;
  return Seconds( ts, 0 );
  }

/*=================================================================== TimeStr
*/
static void TimeStr(  /* Umwandlung einer Sekundenzahl n in eine Zeichen- */
long n,               /* kette t[] der Form  ddd.hh:mm:ss;                */
char *t )             /* fhrende Nullen werden fortgelassen.             */
  {
  int l = 0;
  long k;
  char b[10];
  t[0] = '\0';
  if (n < 0) {  strcat( t, "-" );  n = -n; }
  k = n / 86400L;
  if (k > 0) {
    sprintf( b, "%ld.", k );  strcat( t, b );  n -= k*86400L;  l++; }
  k = n/3600L;
  sprintf(b, "%02ld:", k);
  strcat(t, b);  n -= k*3600L;  l++;
  k = n/60L;
  sprintf(b, "%02ld:", k);
  strcat(t, b);  n -= k*60L; l++;
  sprintf(b, "%02ld", n);
  strcat(t, b);
  return; }

/*=================================================================== Seconds
*/
static long Seconds(  /* Umwandlung einer Zeitangabe t[] von der Form    */
char *t,              /* ddd.hh:mm:ss in die Anzahl der Sekunden. Wenn   */
long r )              /* die Zeichenkette mit '+' beginnt, wird die Se-  */
                      /* kundenzahl zum zweiten Argument r addiert. Zu-  */
                      /* rckgegeben wird die Anzahl der Sekunden.       */
  {
  char *d, s[20];
  long z, x, m;
  strncpy(s, t, 18);  s[18] = 0;
  if (t[0] == 0)  return 0;
  if (t[0] == '+') { s[0] = ' ';  z = r;  m = 1; }
  else {
    z = 0;
    if (t[0] == '-') { s[0] = ' ';  m = -1; }
    else m = 1;
  }
  d = strrchr(s, ':');
  if (d != NULL)  {
    x = atol(d+1);  z +=  m*x;
    *d = 0;
    m *= 60;
    d = strrchr(s, ':');
    if (d != NULL) {
      x = atol(d+1);  z += m*x;
      *d = 0;
      m *= 60;
      d = strrchr(s, '.');
      if (d != NULL) {
        x = atol(d+1);  z += m*x;
        *d = 0;
        m *= 24;
      }
    }
  }
  x = atol(s);  z += m*x;
  return z;
}

//===============================================================================

void DmnReplaceEscape(  // replace ansi escape codes
char *s )               // source and destination string
{
  char *p1, *p2;
  for (p1=s, p2=s; (*p2); ) {
    if (*p2 == '\\') {
      switch (*++p2) {
        case 'b':   *p1 = '\b';  p2++;
                    break;
        case '\\':  *p1 = '\\';  p2++;
                    break;
        case '~':   *p1 = ' ';   p2++;
                    break;
        case 't':   *p1 = '\t';  p2++;
                    break;
        case 'r':   *p1 = '\r';  p2++;
                    break;
        case 'n':   *p1 = '\n';  p2++;
                    break;
        case 'f':   *p1 = '\f';  p2++;
                    break;
        case 'v':   *p1 = '\v';  p2++;
                    break;
        case '\'':  *p1 = '\'';  p2++;
                    break;
        case '\"':  *p1 = '\"';  p2++;
                    break;
        default:    *p1 = '\\';
      }
      p1++;
    }
    else  *p1++ = *p2++;
  }
  *p1 = 0;
}

//====================================================================== AnaForm
static int AnaForm(   // analyse format combination
char *frm,        // extended format
int *psize )      // record size
  {
  dP(AnaForm);
  enum DAT_TYPE dt;
  char *pp, *pa, *pb, *ps;
  char name[40], option[40], lfield[20], lspec, fspec, fm[256];
  int i, n, nn, np, m, l, lf, sz, nform;
  float fac;
  if (FrmTab)  FREE(FrmTab);
  FrmTab = NULL;
  nform = 0;
  pa = frm;
  ps = pa;
  n = 0;
  nn = 0;
  while (NULL != (pp = strchr(pa, '%'))) {
    l = 1;
    if (pp[1] == '[')  sscanf(pp+2, "%d", &l);
    n++;
    nn += l;
    pa = pp + 1;
    }
  pa = frm;
  np = n;
  FrmTab = ALLOC((nn+1)*sizeof(DMNFRMREC));  if (!FrmTab)       eX(1);
  nform = nn;
  nn = 0;
  sz = 0;
  for (n=0; n<np; n++) {
    for (pp=pa; *pp!='%'; pp++)  if (!*pp)                      eX(2);
    *name = 0;
    l = pp - pa;
    if (l > 39) l = 39;
    strncat(name, pa, l);
    m = 1;
    ps = pp;
    if (ps[1] == '[') {
      pb = strchr(ps+2, ']');  if (!pb)                         eX(3);
      if (1 != sscanf(ps+2, "%d", &m))                          eX(4);
      if (m < 0)                                                eX(5);
      ps = pb;
      }
    fac = 0;                                                    //-00-03-10
    *option = 0;
    if (ps[1] == '(') {
      pb = strchr(ps+2, ')');  if (!pb)                         eX(6);
      strncat(option, ps+1, pb-ps);
      if (ps[2] != '*')                                         eX(7);
      if (1 != sscanf(ps+3, "%f", &fac))                        eX(8);
      ps = pb;
      }
    ps++;
    l = strspn(ps, "+-.0123456789");  if (l > 19)               eX(9);
    *lfield = 0;
    lf = 0;
    if (l) {
      strncat(lfield, ps, l);
      if (1 != sscanf(lfield, "%d", &lf))                       eX(10);
      // if (lf < 0)  lf = -lf;                                 -00-03-17
      ps += l;
      }
    if (strchr("hl", *ps)) {
      lspec = *ps;
      ps++;
      }
    else  lspec = 0;
    fspec = *ps;
    if (fspec == 'c')    dt = chDAT;
    else if (strchr("dioux", fspec)) {
      if (lspec == 'h')  dt = siDAT;
      else               dt = inDAT;
      }
    else if (strchr("efg", fspec)) {
      if (lspec == 'l')  dt = lfDAT;
      else               dt = flDAT;
      }
    else if (fspec == 't') {
      if (lspec == 'l')  dt = ltDAT;
      else               dt = tmDAT;
      }
    else                                                        eX(11);
    for (i=0; i<m; i++) {
      l = strlen(name);
      strcpy(FrmTab[nn].name, name);
      FrmTab[nn].dt = dt;
      FrmTab[nn].ld = DatLen[dt];
      FrmTab[nn].lf = lf;
      FrmTab[nn].fac = fac;
      FrmTab[nn].fspc = fspec;
      strcpy(FrmTab[nn].opt, option);
      strcpy(fm, "%");
      strcat(fm, lfield);
      if (fspec == 't')  strcat(fm, "s");
      else {
        strncat(fm, &lspec, 1);
        strncat(fm, &fspec, 1);
      }
      strcpy(FrmTab[nn].oform, fm);
      strcpy(fm, "%");
      strncat(fm, &lspec, 1);
      if (fspec == 't')  strcat(fm, "s");
      else               strncat(fm, &fspec, 1);
      strcpy(FrmTab[nn].iform, fm);
      if (l > 0)  name[l-1]++;
      sz += FrmTab[nn].ld;
      nn++;
      }
    pa = ps + 1;
    }
  if (!*psize)  *psize = sz;
  else if (*psize != sz)                eX(20);
  return nform;
eX_1:
  eMSG("can't allocate format table (%d entries)! (%s)\n", nn, ePGMs);
eX_2: eX_3: eX_4: eX_5: eX_6: eX_7: eX_8: eX_9: eX_10: eX_11:
  vMsg("invalid format string! (%s)\n", ePGMs);
  vMsg("%s\n", frm);
  vMsg("%*s", frm-pa+1, "|");
  if (ps > pa) vMsg("%*s", ps-pa, "^");
  vMsg("\n");
  return MsgCode;
eX_20:
  eMSG("inconsistent element size: format %d, should be %d! (%s)\n",
        sz, *psize, ePGMs);
  }

static int PrnForm( DMNFRMREC *pf, char *s )
  {
  int n;
  if (!pf)  return 0;
  printf("Table of formats %s:\n", s);
  printf("Nr. frm  Name Ld Lf     Factor Output | Input\n");
  for (n=1; (pf->dt); pf++, n++)
    printf("%2d: (%c) %5s %2d %2d %10.3e %s | %s\n", n, pf->fspc, pf->name, pf->ld,
      pf->lf, pf->fac, pf->oform, pf->iform);
  return 0;
  }

static int fPrintR( FILE *f, DMNFRMREC *pf, float fact, void *pr )
{
  int n;
  char c;                                                       //-00-03-17
  float fc;                                                     //-00-03-10
  if (fact <= 0)  fact = 1;                                     //-00-03-10
  n = 0;
  while (pf->dt) {
    if (pf->lf > 0) {                                           //-00-03-17
      fprintf(f, "\t");
      fc = (pf->fac != 0) ? pf->fac : fact;                     //-00-03-10
      switch (pf->dt) {
        case chDAT: c = *(char*)pr;                             //-00-03-17
                    if (c < ' ')  c = ' ';                      //-00-03-17
                    fprintf(f, pf->oform, c);                   //-00-03-17
                    break;
        case siDAT: fprintf(f, pf->oform, *(short int*)pr);
                    break;
        case inDAT: fprintf(f, pf->oform, *(int*)pr);
                    break;
        case flDAT: fprintf(f, pf->oform, fc*(*(float*)pr));    //-00-03-10
                    break;
        case tmDAT: fprintf(f, pf->oform, TmString((long*)pr));
                    break;
        case ltDAT: fprintf(f, pf->oform, MsgDateString(*(double*)pr));
                    break;
        case lfDAT: fprintf(f, pf->oform, fc*(*(double*)pr));   //-00-03-10
                    break;
        default:    ;
      }
      n++;
    }
    pr = (char*)pr + pf->ld;
    pf++;
  }
  return n;
}

/*================================================================ DmnFindName
*/
char *DmnFindName(  // Find a parameter name in the header
char *buf,          // header (lines separated by \n)
char *name )        // parameter name
  {                 // RETURN: pointer to name string
  char *s, *t, white[] = " =\t\n\r";
  int l, i;
  if ((!buf) || (!name))  return NULL;
  l = strlen(name);
  if (!l)  return NULL;
  for (s=buf; (*s); s++) {
    t = s;
    for (i=0; i<l; i++, s++)
      if (tolower(*s) != tolower(name[i]))  break;
    if ((i >= l) && (strchr(white, *s)))  break;
    if (!*s)  return NULL;
    s = strchr(s, '\n');
    if (!s)  return NULL;
    }
  if (!*s)  return NULL;
  return t;
  }

//=============================================================== DmnGetValues
char *DmnGetValues(   // Get the value string of a parameter
char *hdr,            // header
char *name,           // name of the parameter
int *pl )             // set to length of the value string
  {                   // RETURN: pointer to value string
  char *s, *e, white[] = " =\t\n\r";
  *pl = 0;
  s = DmnFindName(hdr, name);
  if (!s)  return NULL;
  s += strlen(name);
  for ( ; (*s); s++)
    if (!strchr(white, *s))  break;
  if (!*s)  return s;
  e = strchr(s, '\n');
  if (!e)  e = s + strlen(s);
  while (*e <= ' ')  e--;
  *pl = e - s + 1;
  return s;
  }

//=============================================================== DmnRplValues
int DmnRplValues(   // Replace the value string of a parameter
TXTSTR *phdr,       // header
char *name,         // name of the parameter
char *values )      // new values (or NULL to delete)
{                   // RETURN: 1 = added, 2 = replaced or deleted
  char *pn, *pc;
  int r = 0;
  TXTSTR newhdr = { NULL, 0 };
  pc = phdr->s;
  pn= DmnFindName(pc, name);
  if (pn) {
    TxtNCat(&newhdr, pc, pn-pc);
    if (values) {
      TxtCat(&newhdr, name);
      TxtCat(&newhdr, "\t");
      TxtCat(&newhdr, values);
    }
    pn = strchr(pn, '\n');
    if (pn)  TxtCat(&newhdr, pn);
    else     TxtCat(&newhdr, "\n");
    TxtClr(phdr);
    *phdr = newhdr;
    r = 2;
  }
  else if (values) {
    TxtCat(phdr, name);
    TxtCat(phdr, "\t");
    TxtCat(phdr, values);
    TxtCat(phdr, "\n");
    r = 1;
  }
  return r;
}

//================================================================= DmnGetCount
int DmnGetCount(  // Get the number of tokens of a parameter
char *hdr,        // header
char *name )      // name of the parameter
{                 // RETURN: number of tokens for this parameter
  int i, l;
  char *s, *pc, *pn;
  TXTSTR t;
  s = DmnGetValues(hdr, name, &l);
  if (!s)  return -1;
  t.s = NULL;
  TxtNCpy(&t, s, l);
  pc = strtok(t.s, SEPCHR);
  for (i=0; ; i++) {
    pc = MsgTok(pn, &pn, SEPCHR, &l);
    if (!pc)  break;
  }
  TxtClr(&t);
  return i;
}

//================================================================= DmnGetInt
int DmnGetInt(  // Get integer values of a parameter
char *hdr,      // header
char *name,     // name of the parameter
char *frm,      // format for reading
int *ii,        // integer vector
int n )         // maximum number of positions in ii[]
{               // RETURN: number of values read
  int i, l;
  char *s, *pc;
  TXTSTR t;
  s = DmnGetValues(hdr, name, &l);
  if (!s)  return -1;
  t.s = NULL;
  TxtNCpy(&t, s, l);
  pc = strtok(t.s, SEPCHR);
  for (i=0; i<n; i++) {
    if (!pc)  break;
    sscanf(pc, frm, ii+i);
    pc = strtok(NULL, SEPCHR);
  }
  TxtClr(&t);
  return i;
}

//================================================================= DmnPutInt
int DmnPutInt(  // Put integer values of a parameter into header
char *hdr,      // header
int len,        // maximum length of header string
char *name,     // name of the parameter
char *frm,      // format for writing (with separator!)
int *ii,        // integer vector
int n )         // number of values to write
{               // RETURN: number of values written
  int i;
  char buf[80];
  if (strlen(hdr)+strlen(name)+1 > len)  return -1;
  strcat(hdr, name);
  for (i=0; i<n; i++) {
    sprintf(buf, frm, ii[i]);
    if (strlen(hdr)+strlen(buf)+1 > len)  return -1;
    strcat(hdr, buf);
  }
  return i;
}

//================================================================= DmnGetFloat
int DmnGetFloat(  // Get float values of a parameter
char *hdr,        // header
char *name,       // name of the parameter
char *frm,        // format for reading
float *ff,        // float vector
int n )           // maximum number of positions in ff[]
{                 // RETURN: number of values read
  int i, l;
  char *s, *pc;
  TXTSTR t;
  s = DmnGetValues(hdr, name, &l);
  if (!s)  return -1;
  t.s = NULL;
  TxtNCpy(&t, s, l);
  pc = strtok(t.s, SEPCHR);
  for (i=0; i<n; i++) {
    if (!pc)  break;
    sscanf(pc, frm, ff+i);
    pc = strtok(NULL, SEPCHR);
  }
  TxtClr(&t);
  return i;
}

//================================================================= DmnPutFloat
int DmnPutFloat(  // Put float values of a parameter into header
char *hdr,        // header
int len,          // maximum length of header string
char *name,       // name of the parameter
char *frm,        // format for writing (with separator!)
float *ff,        // float vector
int n )           // number of values to write
{                 // RETURN: number of values written
  int i;
  char buf[80];
  if (strlen(hdr)+strlen(name)+1 > len)  return -1;
  strcat(hdr, name);
  for (i=0; i<n; i++) {
    sprintf(buf, frm, ff[i]);
    if (strlen(hdr)+strlen(buf)+1 > len)  return -1;
    strcat(hdr, buf);
  }
  return i;
}

//================================================================= DmnGetString
int DmnGetString( // Get string values of a parameter
char *hdr,        // header
char *name,       // name of the parameter
char **ss,        // string vector
int n )           // maximum number of positions in ss[]
{                 // RETURN: number of values read
  int i, l;
  char *s, *pc, *pn;
  TXTSTR t;
  *ss = NULL;
  s = DmnGetValues(hdr, name, &l);
  if (!s)  return -1;
  t.s = NULL;
  TxtNCpy(&t, s, l);
  pn = t.s;
  for (i=0; i<n; i++) {
    pc = MsgTok(pn, &pn, SEPCHR, &l);
    if (!pc)  break;
    ss[i] = TxtAlloc(l+1);
    strncpy(ss[i], pc, l);
  }
  TxtClr(&t);
  return i;
}

//================================================================= DmnPutString
int DmnPutString( // Write string values of a parameter into header
char *hdr,        // header
int len,          // maximum length of header string
char *name,       // name of the parameter
char *frm,        // format for writing (with separator and \"...\" !)
char **ss,        // string vector
int n )           // number of strings to write
{                 // RETURN: number of strings written
  int i;
  char buf[256];
  if (strlen(hdr)+strlen(name)+1 > len)  return -1;
  strcat(hdr, name);
  for (i=0; i<n; i++) {
    sprintf(buf, frm, ss[i]);
    if (strlen(hdr)+strlen(buf)+1 > len)  return -1;
    strcat(hdr, buf);
  }
  return i;
}

//===================================================================== DmnWrite
static int makerecord( char *psrc, char *pdst, DMNFRMREC *pfrm, int swap ) {
  int l, n, i, reverse;
  n = 0;
  while (pfrm->dt) {
    l = pfrm->ld;
    if (pfrm->lf >= 0) {                                            //-00-03-17
      reverse = (swap) && (pfrm->dt!=chDAT);
      if (reverse)
        for (i=0; i<l; i++)  pdst[l-i-1] = psrc[i];
      else memcpy(pdst, psrc, l);
      pdst += l;
      n += l;
    }
    psrc += l;
    pfrm++;
  }
  return n;
}

int DmnWrite(   // Write data into a file
char *fname,    // file name (with path, without extension)
char *hdr,      // header string
ARYDSC *pa )    // address of array descriptor
{
  dP(DmnWrite);
  ARYDSC a, b;
  TXTSTR header = { NULL, 0};
  TXTSTR oform = { NULL, 0 };
  int osize, drop, swap, lsbf, noheader=0, nodata=0;
  FILE *f, *moderr;
  DMNFRMREC *pf;
  char *pb;
  char *mode, *sequ, *form, *data, *pc, *frm, *vldf;
  char fna[256], fnb[256], fn[256], buf[80], format[80], s[80], t[80];
  char oldsel[80], newsel[80], fxd[80], valdef[80];
  int i, j, k, n, binary, reorder, dims, v[ARYMAXDIM], opened, nd, binfile;
  float fact;
  void *pv;
  if (!pa->start)  return 0;
  setmode(2, O_BINARY);
  moderr = fdopen(2, "wb");  if (!moderr)                     eX(0);

  strcpy(fn, fname);
  if (*fn=='-' || *fn=='~') {
    if (fn[1] != 0) {
      noheader = 1;
      nodata = 1;
      if (strchr(fn, 'h'))  noheader = 0;
      if (strchr(fn, 'd'))  nodata = 0;
    }
    strcpy(fn, (*fn=='-') ? "-" : "~");
  }
  TxtCpy(&header, hdr);
  DmnGetString(header.s, "sequ", &sequ, 1);
  DmnGetString(header.s, "mode", &mode, 1);
  DmnGetString(header.s, "data", &data, 1);
  DmnGetString(header.s, "form", &form, 1);
  DmnGetString(header.s, "vldf", &vldf, 1);             //-00-03-03
  if (vldf)  strcpy(valdef, vldf);
  else  *valdef = 0;
  if ((data) && (*data=='*' || *data==0)) {
    FREE(data);
    data = NULL;
  }
  lsbf = 1;
  DmnGetInt(header.s, "lsbf", "%d", &lsbf, 1);
  fact = 0;                                         //-00-03-10
  DmnGetFloat(header.s, "fact", "%f", &fact, 1);
  if (fact <= 0)  fact = 1;                         //-00-03-10
  if (sequ) {
    strcpy(oldsel, sequ);
    AryDefAry(pa, &a, &b, oldsel);                  eG(1);
    strcpy(newsel, oldsel);
  }
  else { a = *pa;  b = a; *newsel = 0; }
  reorder = (a.usrcnt < 0);
  dims = b.numdm;

  *fxd = 0;
  if (sequ) {           // remove fixed dimensions
    strcat(newsel, ",");
    for (n=0; n<dims; n++) {
      t[0] = 'i' + n;
      strcpy(t+1, "=");
      pc = strstr(newsel, t);
      if (pc) {
        for (; (*pc); pc++) {
          if (*pc == ',') { *pc = ' ';   break; }
          *pc = ' ';
        }
        strncat(fxd, t, 1);
      }
    }
    for (i=strlen(fxd)-1; i>=0; i--) {
      for (j=strlen(newsel)-1; j>=0; j--) {
        if (!strchr("ijklm", newsel[j]))  continue;
        if (newsel[j] > fxd[i])  newsel[j]--;
      }
    }
    for (i=0, j=0; i<strlen(newsel); i++)
      if (newsel[i] > ' ')  newsel[j++] = newsel[i];
    newsel[j] = 0;
    if (j > 0)  newsel[j-1] = 0;
  }
  else  strncat(newsel, "i,j,k,l,m", 2*dims-1);
  DmnRplValues(&header, "sequ", TxtQuote(newsel));
  nd = dims - strlen(fxd);

  if (form)  frm = form;
  else {
    sprintf(format, "A%%[%d]13.5e", a.elmsz/sizeof(float));
    frm = format;
  }
  DmnReplaceEscape(frm);
  AnaForm(frm, &a.elmsz);                                           eG(9);
  binary = (mode) ? (0==strcmp(mode, "binary")) : (form == NULL);   //-00-03-17
  drop = 0;
  osize = 0;
  pf = FrmTab;
  TxtCat(&oform, "\"");
  i = 0;
  while (pf->dt) {
    if ((pf->lf>0) || ((binary) && (pf->lf==0))) {                  //-00-03-17
      osize += pf->ld;
      TxtCat(&oform, pf->name);
      TxtCat(&oform, "%");
      TxtCat(&oform, pf->opt);
      TxtCat(&oform, pf->oform+1);
    }
    else {
      drop++;
      if (i < strlen(valdef))  valdef[i] = ' ';                     //-00-03-03
    }
    i++;
    pf++;
  }
  TxtCat(&oform, "\"");
  if (drop) {
    DmnRplValues(&header, "form", oform.s);
    if (vldf) {                                                     //-00-03-03
      for (i=0, k=0; k<strlen(valdef); k++)
        if (valdef[k] != ' ')  valdef[i++] = valdef[k];
      valdef[i] = 0;
      DmnRplValues(&header, "vldf", TxtQuote(valdef));
    }
  }

  sprintf(buf, "dims\t%d\n", nd);       TxtCat(&header, buf);
  sprintf(buf, "size\t%d\n", osize);    TxtCat(&header, buf);
  strcpy(buf, "lowb");
  for (n=0; n<dims; n++) {
    if (strchr(fxd, 'i'+n))  continue;
    sprintf(s, "\t%d", b.bound[n].low);
    strcat(buf, s);
  }
  strcat(buf, "\n");            TxtCat(&header, buf);
  strcpy(buf, "hghb");
  for (n=0; n<dims; n++) {
    if (strchr(fxd, 'i'+n))  continue;
    sprintf(s, "\t%d", b.bound[n].hgh);
    strcat(buf, s);
  }
  strcat(buf, "\n");            TxtCat(&header, buf);

  opened = 0;
  binfile = 0;
  *fna = 0;
  *fnb = 0;
  if (!strncmp(fn, "LOG", 3)) {
    if (binary)  goto done;
    if (!MsgFile)  goto done;
    if (data)  goto done;
    f = MsgFile;
    fprintf(f, "%s", fn+3);
    }
  else if (!strcmp(fn,"stdout") || !strcmp(fn,"stderr") || *fn=='-' || *fn=='~') {
    f = (!strcmp(fn,"stderr")) ? stderr : stdout;
    if (!noheader) {
      if (1 != fwrite(header.s, header.l, 1, f))      eX(3);
    }
    fprintf(f, "*\n");
    if (*fn == '~')  f = moderr;
    }
  else {
    strcpy(fna, fn);
    for (pc=fna; (*pc); pc++)  if (*pc == '\\')  *pc = '/';
    strcpy(fnb, fna);
    strcat(fna, STDEXT);
    f = fopen(fna, "wb");  if (!f)              eX(2);
    opened = 1;
    if (1 != fwrite(header.s, header.l, 1, f))        eX(3);
    fprintf(f, "*\n");
    if (binary)  binfile = 1;
    }
  TxtClr(&header);
  if (data || binfile) {
    if (opened)  fclose(f);
    if (data) {
        pc = strrchr(fnb, '/');
        if (pc)  pc++;
      else  pc = fnb;
        *pc = 0;
        strcat(fnb, data);
      }
      strcat(fnb, (binary) ? BINEXT : TXTEXT);
      f = fopen(fnb, "wb");  if (!f)              eX(4);
    opened = 1;
  }

  if (nodata)  goto done;
  if (binary) {   // unformatted
    n = 1;
    pb = (char*) &n;
    swap = (pb[0] != lsbf);
    if (drop || reorder || swap) {                                  //-00-03-17
      pb = ALLOC(a.elmsz);
      for (n=0; n<dims; n++)  v[n] = a.bound[n].low;
      do {
        pc = AryPtr(&a, v[0], v[1], v[2], v[3], v[4]);  if (!pc)    eX(5);
        if (drop || swap) {
          n = makerecord(pc, pb, FrmTab, swap);
          if (1 != fwrite(pb, n, 1, f))                             eX(6);
        }
        else {
          if (1 != fwrite(pc, a.elmsz, 1, f))                       eX(6);
        }
        for (n=a.numdm-1; n>=0; n--) {
          v[n]++;
          if (v[n]>a.bound[n].hgh && n>0)  v[n] = a.bound[n].low;
          else  break;
        }
        fflush(moderr);
      } while (v[0] <= a.bound[0].hgh);
      FREE(pb);
    }
    else {
      if (1 != fwrite(a.start, a.ttlsz, 1, f))                      eX(7);
    }
  }
  else {  // formatted
    for (n=0; n<dims; n++)  v[n] = a.bound[n].low;
    do {
      pv = AryPtr(&a, v[0], v[1], v[2], v[3], v[4]);  if (!pv)      eX(8);
      if (1 < fPrintR(f, FrmTab, fact, pv))  fprintf(f, "\n");
      for (n=a.numdm-1; n>=0; n--) {
        v[n]++;
        if (v[n]>a.bound[n].hgh && n>0) {
          v[n] = a.bound[n].low;
          fprintf(f, "\n");
        }
        else  break;
      }
    } while (v[0] <= a.bound[0].hgh);
    fprintf(f, "\n");
    if (opened)  fprintf(f, "***\n");
  }
  if (opened)  fclose(f);
  // fclose(moderr);
done:
  if (mode)  FREE(mode);
  if (data)  FREE(data);
  if (form)  FREE(form);
  if (sequ)  FREE(sequ);
  return 1;
eX_0:
  eMSG("can't reopen stderr! (%s)\n", ePGMs);
eX_1:
  eMSG("can't get subarray with sequence \"%s\"! (%s)\n", sequ, ePGMs);
eX_2:
  eMSG("can't open descriptor file %s! (%s)\n", fna, ePGMs);
eX_3:
  eMSG("can't write header to file %s! (%s)\n", fna, ePGMs);
eX_4:
  eMSG("can't open data file %s! (%s)\n", fnb, ePGMs);
eX_5: eX_8:
  eMSG("index error writing to file %s! (%s)\n", fnb, ePGMs);
eX_6: eX_7:
  eMSG("can't write data to file %s! (%s)\n", fnb, ePGMs);
eX_9:
  eMSG("invalid format %s! (%s)\n", frm, ePGMs);
}


//===================================================================== DmnRead
int DmnRead(    // Read data from a file
char *fn,       // file name (with path, without extension)
TXTSTR *pu,     // user header text
TXTSTR *ps,     // system header text
ARYDSC *pa )    // address of array descriptor
{
  dP(DmnRead);
  ARYDSC a, b;
  FILE *f;
  char *mode, *sequ, *form, *data, *pc, *pn, *buf, *frm;
  char fna[256], fnb[256], name[80], format[32];
  int n, nn, l, binary, reorder, dims, v[ARYMAXDIM], bufsize, size;
  int ii[2][ARYMAXDIM];
  float fact, fc;
  DMNFRMREC *pf;
  void *pv;
  union {
    char c;
    short int h;
    int i;
    long l;
    float f;
    double d;
  } u;
  TXTSTR syshdr = { NULL, 0 };
  TXTSTR usrhdr = { NULL, 0 };
  if (!ps)  ps = &syshdr;
  if (!pu)  pu = &usrhdr;
  if (pa->start)  AryFree(pa);                                eG(1);
  if (!strcmp(fn, "-") || !strcmp(fn, "stdin"))  f = stdin;
  else {
    strcpy(fna, fn);
    for (pc=fna; (*pc); pc++)  if (*pc == '\\')  *pc = '/';
    strcpy(fnb, fna);
    strcat(fna, STDEXT);
    f = fopen(fna, "rb");  if (!f)                            eX(2);
  }
  TxtClr(pu);
  TxtClr(ps);
  bufsize = 4000;
  buf = TxtAlloc(bufsize);
  nn = 0;
  while (fgets(buf, bufsize, f)) {  // read header
    nn++;
    if (*buf == '*')  break;
    *name = 0;
    strncat(name, buf, 4);
    MsgLow(name);
    if (strstr(DimParms, name)) TxtCat(ps, buf);
    else                        TxtCat(pu, buf);
    pc = DmnFindName(buf, "buff");
    if (pc) {
      n = bufsize;
      sscanf(pc, "%d", &n);
      if (n > bufsize) {      // resize input buffer
        FREE(buf);
        bufsize = n;
        buf = TxtAlloc(bufsize);
      }
    }
  }
  DmnGetString(ps->s, "sequ", &sequ, 1);
  DmnGetString(ps->s, "mode", &mode, 1);
  DmnGetString(ps->s, "data", &data, 1);
  DmnGetString(ps->s, "form", &form, 1);
  binary = (mode) ? (0==strcmp(mode, "binary")) : (form == NULL); //-00-03-17
  size = 0;                                                       //-00-03-11
  DmnGetInt(ps->s, "size", "%d", &size, 1);                       //-00-03-11
  if ((data) && (*data=='*' || *data==0)) {
    FREE(data);
    data = NULL;
  }
  if ((binary) && (size<=0))                                      eX(16);
  if (form)  frm = form;
  else {
    if (size >= sizeof(float))  sprintf(format, "%%[%d]f", size/sizeof(float));
    else                                                          eX(9);
    frm = format;                                                 //-00-03-11
  }
  AnaForm(frm, &size);                                            eG(9);
  fact = 0;                                                       //-00-03-10
  DmnGetFloat(ps->s, "fact", "%f", &fact, 1);
  if (fact <= 0)  fact = 1;                                       //-00-03-10
  if (1 != DmnGetInt(ps->s, "dims", "%d", &dims, 1))              eX(17);
  if (dims<1 || dims>ARYMAXDIM)                                   eX(10);
  if (dims != DmnGetInt(ps->s, "lowb", "%d", ii[0], ARYMAXDIM))   eX(11);
  if (dims != DmnGetInt(ps->s, "hghb", "%d", ii[1], ARYMAXDIM))   eX(11);
  AryCreate(pa, size, dims, ii[0][0], ii[1][0], ii[0][1], ii[1][1],
    ii[0][2], ii[1][2], ii[0][3], ii[1][3], ii[0][4], ii[1][4]);  eG(12);
  if (sequ) {
    AryDefAry(pa, &a, &b, sequ);                                  eG(13);
  }
  else { a = *pa;  b = a; }
  reorder = (a.usrcnt < 0);
  dims = b.numdm;
  if ((binary || (data)) && (f!=stdin)) {
    fclose(f);
    if (data) {
      pc = strrchr(fnb, '/');
      if (pc)  pc++;
      else  pc = fnb;
      *pc = 0;
      strcat(fnb, data);
    }
    strcat(fnb, (binary) ? BINEXT : TXTEXT);
    f = fopen(fnb, "rb");  if (!f)                                eX(4);
    nn = 0;
  }
  if (binary) {   // unformatted
    if (reorder) {
      for (n=0; n<dims; n++)  v[n] = a.bound[n].low;
      do {
        pc = AryPtr(&a, v[0], v[1], v[2], v[3], v[4]);  if (!pc)  eX(5);
        if (1 != fread(pc, a.elmsz, 1, f))                        eX(6);
        for (n=a.numdm-1; n>=0; n--) {
          v[n]++;
          if (v[n]>a.bound[n].hgh && n>0)  v[n] = a.bound[n].low;
          else  break;
        }
      } while (v[0] <= a.bound[0].hgh);
    }
    else {
      if (1 != fread(a.start, a.ttlsz, 1, f))                     eX(7);
    }
  }
  else {  // formatted
    for (n=0; n<dims; n++)  v[n] = a.bound[n].low;
    *buf = 0;
    pn = buf;
    do {
      pv = AryPtr(&a, v[0], v[1], v[2], v[3], v[4]);  if (!pv)    eX(8);
      pf = FrmTab;
      while (pf->dt) {
        while (NULL == (pc=MsgTok(pn,&pn," ;\t\r\n",&l))) {
          if (!fgets(buf, bufsize, f))                            eX(14);
          nn++;
          pn = buf;
        }
        if      (pf->dt == ltDAT)  u.d = MsgDateVal(pc);
        else if (pf->dt == tmDAT)  u.l = TmValue(pc);
        else if (1 != sscanf(pc, pf->iform, &u))            eX(15);
        fc = (pf->fac != 0) ? pf->fac : fact;               //-00-03-10
        switch (pf->dt) {
          case chDAT: *(char*)pv = u.c;
                      break;
          case siDAT: *(short int*)pv = u.h;
                      break;
          case inDAT: *(int*)pv = u.i;
                      break;
          case flDAT: *(float*)pv = u.f/fc;                 //-00-03-10
                      break;
          case tmDAT: *(long*)pv = u.l;
                      break;
          case ltDAT: *(double*)pv = u.d;
                      break;
          case lfDAT: *(double*)pv = u.d/fc;                //-00-03-10
                      break;
          default:    ;
        }
        pv = (char*)pv + pf->ld;
        pf++;
      }
      for (n=a.numdm-1; n>=0; n--) {
        v[n]++;
        if (v[n]>a.bound[n].hgh && n>0) {
          v[n] = a.bound[n].low;
        }
        else  break;
      }
    } while (v[0] <= a.bound[0].hgh);
  }
  if (f != stdin)  fclose(f);
  TxtClr(&usrhdr);
  TxtClr(&syshdr);
  if (mode)  FREE(mode);
  if (data)  FREE(data);
  if (form)  FREE(form);
  if (sequ)  FREE(sequ);
  return 1;
eX_1:
  eMSG("internal error! (%s)\n", ePGMs);
eX_2:
  eMSG("can't open descriptor file %s! (%s)\n", fna, ePGMs);
eX_4:
  eMSG("can't open data file %s! (%s)\n", fnb, ePGMs);
eX_5: eX_8:
  eMSG("index error writing to file %s! (%s)\n", fnb, ePGMs);
eX_6: eX_7:
  eMSG("can't write data to file %s! (%s)\n", fnb, ePGMs);
eX_9:
  eMSG("invalid format specification %s! (%s)\n", frm, ePGMs);
eX_10:
  eMSG("invalid number (%d) of dimensions in file %s! (%s)\n", dims, fnb, ePGMs);
eX_11:
  eMSG("missing bounds for file %s! (%s)\n", fnb, ePGMs);
eX_12:
  eMSG("can't create array for file %s! (%s)\n", fnb, ePGMs);
eX_13:
  eMSG("invalid selection %s for file %s! (%s)\n", sequ, fnb, ePGMs);
eX_14:
  eMSG("unexpected end of file %s! (%s)\n", fnb, ePGMs);
eX_15:
  eMSG("can't read item in line %d, file %s! (%s)\n", nn, fnb, ePGMs);
eX_16:
  eMSG("missing parameter \"size\" in file %s! (%s)\n", fna, ePGMs);
eX_17:
  eMSG("missing parameter \"dims\" in file %s! (%s)\n", fna, ePGMs);
}

#ifdef MAIN /*###############################################################*/

#define  ISOPTION(a)  (NULL != strchr(SYMOPT, (a)))

static ARYDSC Ary;
static char Path[256];

void Help( void )
  {
  printf("IBJdata:  Input and Output of data arrays, version %1.2f\n",VERSION);
  printf("usage:    IBJdata <path> <option ...>\n");
  printf("<path>    working directory\n");
  printf("<option>  -a<factor>   : scaling factor\n");
  printf("          -d<data>     : name of output data file\n");
  printf("          -f<format>   : format of output data\n");
  printf("          -i<input>    : name of input file\n");
  printf("          -m<mode>     : \"binary\" or \"text\"\n");
  printf("          -n<name>     : name of ouput descriptor file\n");
  printf("          -p<print>    : print output on screen\n");
  printf("          -s<sequence> : index order in output data\n");
  exit(0);
  }

int main( int argc, char *argv[] )
{
  dP(DmnMain);
  int i, j, k, n, printing, list_format;
  float fact;
  char s[80], *pc;
  char form[400], onam[MAXFNLEN], sequ[80], data[MAXFNLEN];
  char mode[80], inam[MAXFNLEN], fn[MAXFNLEN];
  TXTSTR usrhdr = { NULL, 0 };
  TXTSTR syshdr = { NULL, 0 };
  if (argc < 2)  Help();
  strcpy(onam, "test");
  *inam = 0;
  *data = 0;
  *form = 0;
  *mode = 0;
  *sequ = 0;
  printing = 0;
  list_format = 0;
  fact = 0;
  for (n=1; n<argc; n++) {
    strcpy(s, argv[n]);
    if (ISOPTION(*s)) {
      switch (s[1]) {
        case 'a': sscanf(s+2, "%f", &fact);
              break;
        case 'd': strcpy(data, s+2);
              break;
        case 'f': strcpy(form, s+2);
              break;
        case 'i': strcpy(inam, s+2);
              break;
        case 'l': list_format = 1;
              break;
        case 'm': strcpy(mode, s+2);
              break;
        case 'o': strcpy(onam, s+2);
              break;
        case 'p': printing = 1;
              break;
        case 's': strcpy(sequ, s+2);
              break;
        default:  ;
      }
    }
    else  strcpy(Path, s);
  }
  if (*Path) {
    for (pc=Path; (*pc); pc++)  if (*pc == '\\')  *pc = '/';
    if (pc[-1] != '/') {
      pc[0] = '/';
      pc[1] = 0;
    }
  }
  else strcpy(Path, "./");
  strcpy(fn, Path);
  strcat(fn, "data.log");
  MsgFile = fopen(fn, "wb");
  if (!MsgFile) {
    printf("can't open log-file!\n");
    exit(1);
  }
  MsgVerbose = 1;
  vMsg("IBJdata: Input and Output of data arrays, version %1.2f\n", VERSION);
    vMsg("Copyright Dr. Lutz Janicke, Dunum, Germany, 1998\n");
    vMsg("Compiled at %s  %s\n", __DATE__, __TIME__);
  for (i=0; i<n; i++)  fprintf(MsgFile, ">%s\n", argv[i]);
  if (*inam) {
    strcpy(fn, Path);
    strcat(fn, inam);
    DmnRead(fn, &usrhdr, &syshdr, &Ary);            eG(0);
  }
  else {
      AryCreate(&Ary, sizeof(float), 3, 1, 3, 2, 4, 0, 4);    eG(1);
      for (i=1; i<=3; i++)
      for (j=2; j<=4; j++)
        for (k=0; k<=4; k++)
          *(float*)AryPtrX(&Ary, i, j, k) = 100*i + 10*j + k;
  }
  if (list_format)  PrnForm(FrmTab, "in input file");
  strcpy(fn, Path);
  strcat(fn, onam);
  if (*mode) {
    sprintf(s, "mode\t%s\n", mode); TxtCat(&usrhdr, s);
  }
  if (*data) {
    sprintf(s, "data\t%s\n", data); TxtCat(&usrhdr, s);
  }
  if (*sequ) {
    sprintf(s, "sequ\t%s\n", sequ); TxtCat(&usrhdr, s);
  }
  if (*form) {
    sprintf(s, "form\t%s\n", form); TxtCat(&usrhdr, s);
  }
  if (fact > 0) {
    sprintf(s, "fact\t%10.3e\n", fact); TxtCat(&usrhdr, s);
  }
  vMsg("OUT:\n%s", usrhdr.s);
  DmnWrite(fn, usrhdr.s, &Ary);             eG(2);
  if (list_format)  PrnForm(FrmTab, "in output file");
  vMsg("finished\n");
    return 0;
eX_0:
    eMSG("no input file! (%s)\n", ePGMs);
eX_1:
    eMSG("can't create data! (%s)\n", ePGMs);
eX_2:
    eMSG("can't write file %s! (%s)\n", onam, ePGMs);
}

#endif  /*###################################################################*/

